﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;

namespace ZavrsniRad
{
    class NeuralNetwork
    {
        #region varijable

        private List<BitArray[][]> _chromosome;

        public List<BitArray[][]> Chromosome
        {
            get { return _chromosome; }
            set { _chromosome = value; }
        }

        private List<double[][]> _weightMatrix;

        public List<double[][]> WeightMatrix
        {
            get { return _weightMatrix; }
            set { _weightMatrix = value; }
        }

        private List<int> _matrixDesign;

        public List<int> MatrixDesign
        {
            get { return _matrixDesign; }
            set { _matrixDesign = value; }
        }

        #endregion

        public NeuralNetwork(List<BitArray[][]> chromosome, List<int> matrixDesign, double minimalWeight, double maximalWeight)
        {
            this.Chromosome = chromosome;
            this.MatrixDesign = matrixDesign;
            UpdateWeightMatrix(this.Chromosome, minimalWeight, maximalWeight);
        }
        /// <summary>
        ///  Metoda računa izlaze iz neuronske mreže, pri računanju se koristi weightMatrix
        /// </summary>
        /// <param name="inputs"></param>
        /// <returns></returns>
        public List<double> CalculateOutput(List<double> inputs)
        {
            List<double> result = new List<double>();
            List<double> intermediate = new List<double>();
            List<double> subMatrixResult = new List<double>();
            for (int i = 0; i < inputs.Count(); i++) subMatrixResult.Add(inputs[i]); //inicjalizacija ulaza

            //prolazimo kroz sve matrice, čuvamo međurezultate, ako je NN definirana kao 3x3x2 imamo 2 matrice: 3x3 i 3x2
            for (int i = 0; i < _weightMatrix.Count; i++) 
            {
                //uzimamo matricu iz liste matrica
                double[][] subMatrixToCompute = _weightMatrix[i];
                intermediate.Clear();
                //računamo izlaze na kraju odabrane matrice, koji će postati ulazi sljedeće matrice, ili u konačnici output matrice
                for (int j = 0; j < this.MatrixDesign[i+1]; j++)
                {
                    intermediate.Add(0);
                    for (int k = 0; k < this.MatrixDesign[i]; k++)
                    {
                        intermediate[j] += subMatrixResult[k] * subMatrixToCompute[k][j];      
                    }
                }
                subMatrixResult.Clear();
                if (i < (_weightMatrix.Count - 1))
                {
                    for (int j = 0; j < intermediate.Count(); j++)
                    {
                        subMatrixResult.Add(Math.Tanh(intermediate[j])); 
                    }
                }
                else
                {
                    for (int j = 0; j < intermediate.Count(); j++)
                    {
                        subMatrixResult.Add(1 / (1 + Math.Exp((-1) * intermediate[j])));
                    }
                }
            }
            return subMatrixResult;
        }
   
        #region Pomoćne metode
        private void UpdateWeightMatrix(List<BitArray[][]> chromosome, double minWeight, double maxWeight)
        {
            this.WeightMatrix = new List<double[][]>();
            for (int i = 0; i < this.MatrixDesign.Count() - 1; i++)
            {
                double[][] matrixToAdd = new double[this.MatrixDesign[i]][];
                BitArray[][] chromosomeMatrix = chromosome[i];
                for (int j = 0; j < this.MatrixDesign[i]; j++)
                {
                    matrixToAdd[j] = new double[this.MatrixDesign[i + 1]];

                    for (int k = 0; k < this.MatrixDesign[i + 1]; k++)
                    {
                        matrixToAdd[j][k] = BitArrayToWeight(chromosomeMatrix[j][k], minWeight, maxWeight);
                    }
                }
                this.WeightMatrix.Add(matrixToAdd);
            }

        }

        private int BitArrayToInt32(BitArray bitArray)
        {
            if (bitArray.Count <= 32)
            {
                int[] result = new int[1];
                bitArray.CopyTo(result, 0);
                return result[0];
            }
            else return 0;
        }

        private double BitArrayToWeight(BitArray bitArray, double minValue, double maxValue)
        {
            if (bitArray[bitArray.Count-1] == true) //zadnji bit predstavlja negativan predznak TRUE - negativan
            {
                return ((double)(BitArrayToInt32(bitArray)) / Math.Pow(2, bitArray.Count - 1)) * minValue;
            }
            else
            {
                return ((double)(BitArrayToInt32(bitArray)) / Math.Pow(2, bitArray.Count - 1)) * maxValue;
            }
        }
        #endregion
    }
}
